#include "syncdata.h"
namespace Localization
{
namespace PointLoc
{

SyncData::SyncData() {
  std::cout << "syncdata_构造完成，开启功能请调用Init()函数" << std::endl;
}

SyncData::~SyncData() {

}

void SyncData::Init() {

  std::thread syncTask{std::thread([this] { SyncThread(); })};
  pthread_setname_np(syncTask.native_handle(), "syncTask");
  syncTask.detach();
}

void SyncData::SyncThread() {
  while (true) {
    std::deque<J2Data> j2MsgBuffer;
    std::unique_lock<std::mutex> lk(cammtx_);
    con_.wait(lk, [&] {
      j2MsgBuffer = j2MsgBuffer_;
      j2MsgBuffer_.clear();
      return !j2MsgBuffer.empty();
    });
    lk.unlock();
    while (!j2MsgBuffer.empty()) {
      j2Msg_ = j2MsgBuffer.front();
      j2Inited_ = true;
      j2MsgBuffer.pop_front();
      MatchFrame mixData;
      if (!GetFrame(mixData)) {
        continue;
      }
      sycmtx_.lock();
      synced_.push_back(mixData);
      sycmtx_.unlock();
    }
  }
}

void SyncData::SetInput(MapSwitched &mapLaneMsg) {
  mapmtx_.lock();
  mapMsgBuffer_.push_back(mapLaneMsg);
  mapInited_ = true;
  if (mapMsgBuffer_.size() > 10) {
    mapMsgBuffer_.pop_front();
  }
  mapmtx_.unlock();
}

void SyncData::SetEgoInput(Localization::PointLoc::InsInfo &egoInfo) {
  egomtx_.lock();
  egoMsgBuffer_.push_back(egoInfo);
  egoInited_ = true;
  if (egoMsgBuffer_.size() > 50) {
    egoMsgBuffer_.pop_front();
  }
  egomtx_.unlock();
}

void SyncData::SetInput(J2Data &camLaneMsg) {
  cammtx_.lock();
  j2MsgBuffer_.push_back(camLaneMsg);
  if (j2MsgBuffer_.size() > 10) {
    j2MsgBuffer_.pop_front();
  }
  cammtx_.unlock();
  con_.notify_one();
}

void SyncData::SwitchMapLaneInfo(
    const MapSwitched &mapLaneRaw, Localization::PointLoc::InsInfo &basePose,
    std::vector<Eigen::Vector3d> &mapPtsUTM,
    std::vector<Eigen::Vector3d> &mapPtsVeh,
    std::vector<std::pair<int, std::vector<Eigen::Vector3d>>> &mapLaneswithid,
    std::vector<double> &mapc1s, std::vector<double> &mapc0s) {
  mapc1s.clear();
  mapc0s.clear();
  auto mapLaneNum = mapLaneRaw.mapLinesUTM.size();
  for (int i = 0; i < mapLaneNum; i++) {
    if (i == 0 || i == 3 || i == 4) {
      continue;
    }
    auto oneLine = mapLaneRaw.mapLinesUTM[i];
    auto mapPtsNum = oneLine.size();
    std::vector<Eigen::Vector3d> slctedVeh, slctedUTM;

    for (int j = 0; j < mapPtsNum; j++) {
      // std::cout << "latitude zone: " << laneLocUtm.latitude_zone
      //           << ", Longtitude: " << laneLocUtm.longitude_zone_number
      //           << std::endl;
      Eigen::Vector3d ptUtM = oneLine[j];
      Eigen::Vector3d ptVeh =
        basePose.quat_.inverse() * (ptUtM - basePose.utmPosition_);
      // std::cout << "转到自车坐标系下的点:" << ptVeh.transpose() << std::endl;
      if (ptVeh.x() < 150 && ptVeh.x() > -150 && fabs(ptVeh.y()) < 40) {
        slctedVeh.emplace_back(ptVeh);
      }
    }
    if (isInterpolate_ && slctedVeh.size() >= 2) {
      std::vector<Eigen::Vector3d> resfull;
      for (int k = 0; k < slctedVeh.size() - 1; k++) {
        std::vector<Eigen::Vector3d> res;
        Interpolate3DBetweenTwoPointsByMeter(slctedVeh[k], slctedVeh[k + 1], 5,
                                             res);
        if (res.size() > 0)
          resfull.insert(resfull.end(), res.begin(), res.end());
      }
      resfull.emplace_back(slctedVeh.back());
      if (resfull.size() >= slctedVeh.size())
        slctedVeh = resfull;
    }
    mapPtsVeh.insert(mapPtsVeh.end(), slctedVeh.begin(), slctedVeh.end());
    for (auto &ptVeh : slctedVeh) {
      mapPtsUTM.emplace_back(basePose.quat_ * ptVeh + basePose.utmPosition_);
    }
    if (slctedVeh.size() >= 2) {
      mapLaneswithid.emplace_back(
          std::pair<int, std::vector<Eigen::Vector3d>>(i, slctedVeh));
    }
    slctedVeh.clear();
  }

// std::cout << "Mapc0s size: " << mapLanes.size() << std::endl;
  std::vector<std::pair<int, std::vector<Eigen::Vector3d>>> mapLaneswithidPro;

  for (auto linewithid : mapLaneswithid) {
    auto lower =
        std::lower_bound(linewithid.second.begin(), linewithid.second.end(), 0,
                         comp<Eigen::Vector3d>);

    if (lower == linewithid.second.end() || lower < linewithid.second.begin()) {
      continue;
    }

    Eigen::Vector2d xvals, yvals;
    xvals << lower->x(), (lower - 1)->x();
    yvals << lower->y(), (lower - 1)->y();

    Eigen::Vector2d poly = PolyFit(xvals, yvals);
    // auto mapLane = linewithid.second;
    mapc1s.emplace_back(poly.y());
    mapc0s.emplace_back(poly.x());
    mapLaneswithidPro.emplace_back(linewithid);
  }
  mapLaneswithid = mapLaneswithidPro;

  if (mapc1s.size() == 0) {
    std::cerr << "No Map Lane statisfy" << std::endl;
    return;
  }

  Eigen::Map<Eigen::VectorXd> c1Mat(mapc1s.data(), mapc1s.size());

  Eigen::MatrixXd::Index p, q;
  auto c1Sub =
      (c1Mat - (Eigen::VectorXd::Ones(c1Mat.size()) * c1Mat.mean())).cwiseAbs();
  auto maxdiff = c1Sub.maxCoeff(&p, &q);
  auto mindiff = c1Sub.minCoeff();

  if (maxdiff > mindiff + 0.2) {
    // std::cout << "-------- Not Parallel --------"
    //           << "\n";
    // std::cout << "min diff: " << mindiff << "\n";
    // std::cout << "max diff: " << maxdiff << "\n";
    // std::cout << "max diff index: " << p << std::endl;
    mapLaneswithid.erase(mapLaneswithid.begin() + p);
    mapc1s.erase(mapc1s.begin() + p);
    mapc0s.erase(mapc0s.begin() + p);
  }
}

Eigen::Vector2d SyncData::PolyFit(Eigen::Vector2d &xvals, Eigen::Vector2d &yvals) {
  Eigen::MatrixXd A(xvals.size(), 2);
  for (int i = 0; i < xvals.size(); i++) {
    A(i, 0) = 1.0;
  }
  A(0, 1) = A(0, 0) * xvals(0);
  A(1, 1) = A(1, 0) * xvals(1);
  auto Q = A.householderQr();
  auto result = Q.solve(yvals);
  return result;
}

/// @brief SwitchCamLaneInfoByMap
/// @param camLaneRaw input: camRawdata
/// @param mapLanes input: mapLanes
/// @param mapc1s input: mapc1s
/// @param mapPtsVeh input: mapPtsVeh
/// @param basePose input: NearestEgo
/// @param j2PtsVehDraw output: camLane to draw(Veh)
/// @param j2PtsUTMDraw output: camLane to draw(UTM)
/// @param mapPtsVehMatch output:
/// @param mapPtsUTMMatch output:
/// @param j2PtsVehMatch output:
/// @param j2PtsUTMMatch output:
/// @param camvaild output:
/// @param mapvaild output:
/// @param matches output: matches (camid : mapid)(total 4: total 7)
void SyncData::SwitchCamLaneInfoByMap(
    J2Data &camLaneRaw,
    std::vector<std::pair<int, std::vector<Eigen::Vector3d>>> &mapLanes,
    std::vector<double> &mapc1s, std::vector<double> &mapc0s,
    std::vector<Eigen::Vector3d> &mapPtsVeh,
    Localization::PointLoc::InsInfo &basePose,
    std::vector<Eigen::Vector3d> &j2PtsVehDraw,
    std::vector<Eigen::Vector3d> &j2PtsUTMDraw,
    std::vector<Eigen::Vector3d> &mapPtsVehMatch,
    std::vector<Eigen::Vector3d> &mapPtsUTMMatch,
    std::vector<Eigen::Vector3d> &j2PtsVehMatch,
    std::vector<Eigen::Vector3d> &j2PtsUTMMatch,
    std::vector<std::pair<int, std::vector<Eigen::Vector3d>>> &camvaild,
    std::vector<std::pair<int, std::vector<Eigen::Vector3d>>> &mapvaild,
    std::vector<std::pair<int, int>> &matches) {

  std::unordered_map<int, int> idmap;
  for (int i = 0; i < mapLanes.size(); i++) {
    idmap[mapLanes[i].first] = i;
  }

  // std::cout << "begin Switch cam lane" << std::endl;

  J2Data camLaneProed = camLaneRaw;
  std::vector<Eigen::Vector3d> mapPtsUTMMatchc, mapPtsVehMatchc, j2PtsUTMMatchc,j2PtsVehMatchc;
  Eigen::Vector4d camc0s;
  camc0s[0] = camLaneRaw.lane_markers[0].c0;
  camc0s[1] = camLaneRaw.lane_markers[1].c0;
  camc0s[2] = camLaneRaw.lane_markers[2].c0;
  camc0s[3] = camLaneRaw.lane_markers[3].c0;
  Eigen::Vector4d camc1s;
  camc1s[0] = camLaneRaw.lane_markers[0].c1;
  camc1s[1] = camLaneRaw.lane_markers[1].c1;
  camc1s[2] = camLaneRaw.lane_markers[2].c1;
  camc1s[3] = camLaneRaw.lane_markers[3].c1;

  double disLLCam = fabs(camc0s[2] - camc0s[0]);
  double disRRCam = fabs(camc0s[3] - camc0s[1]);
  if (disLLCam < 1.2) {
    std::cout << "ll too close" << std::endl;
    camc0s[2] = 1e10;
    if (idmap.find(1) != idmap.end() && idmap.find(5) != idmap.end()) {
      double disLLMap = fabs(mapc0s[idmap[1]] - mapc0s[idmap[5]]);
      if (disLLMap < 1.2) {
        mapc0s.erase(mapc0s.begin() + idmap[5]);
        mapc1s.erase(mapc1s.begin() + idmap[5]);
        mapLanes.erase(mapLanes.begin() + idmap[5]);
      }
    }
  }

  idmap.clear();
  for (int i = 0; i < mapLanes.size(); i++) {
    idmap[mapLanes[i].first] = i;
  }

  if (disRRCam < 1.2) {
    std::cout << "rr too close" << std::endl;
    camc0s[3] = 1e10;
    if (idmap.find(2) != idmap.end() && idmap.find(6) != idmap.end()) {
      double disRRMap = fabs(mapc0s[idmap[2]] - mapc0s[idmap[6]]);
      if (disRRMap < 1.2) {
        mapc0s.erase(mapc0s.begin() + idmap[6]);
        mapc1s.erase(mapc1s.begin() + idmap[6]);
        mapLanes.erase(mapLanes.begin() + idmap[6]);
      }
    }
  }

  std::vector<std::pair<int, int>> cam2mapMatch;
  std::unordered_map<int, std::vector<int>> c2mmatchc0;
  int count = 0;

  for (int i = 0; i < mapLanes.size(); i++) {
    double fy = mapc0s[i];
    std::cout << "mapLine id: " << mapLanes[i].first << ", mapLine c0: " << fy
              << std::endl;

    Eigen::Vector4d camc0ssub =
      (camc0s - Eigen::Vector4d::Ones() * fy).cwiseAbs();
    Eigen::MatrixXd::Index p, q;
    auto mindiff = camc0ssub.minCoeff(&p, &q);
    if (mindiff > 3) {
      // std::cout << "wrong match: " << p << " " << i << "diff: " << mindiff
      //           << std::endl;
      // std::cout << fy << "," << fPt.x() << std::endl;
      continue;
    }
    c2mmatchc0[p].emplace_back(i);
    count++;
  }
  if (c2mmatchc0.size() == 0) {
    std::cout << "no match" << std::endl;
    return;
  }

  for (auto match : c2mmatchc0) {
    if (match.second.size() > 1) {
      std::vector<double> tmpSum, tmpVar;
      for (int &mc : match.second) {
        auto &camLine = camLaneRaw.lane_markers[match.first];
        auto &mapLine = mapLanes[mc].second;
        std::vector<double> container;
        for (int i = 0; i < mapLine.size(); i++) {
          auto &mapPtVeh = mapLine[i];
          double x = mapPtVeh.x();
          double y = camLine.c0 + camLine.c1 * x + camLine.c2 * x * x +
                     camLine.c3 * x * x * x;
          // if (minPt.x() > start && minPt.x() < 20 && minPt.x() < end) {
          if (x <= camLine.end && x >= camLine.start) {
            container.emplace_back(fabs(y - mapPtVeh.y()));
          }
        }
        if (container.empty()) {
          tmpVar.emplace_back(1e10);
          tmpSum.emplace_back(1e10);
          continue;
        }
        Eigen::Map<Eigen::VectorXd> containerMat(container.data(),
                                                 container.size());
        double mean = containerMat.mean();
        auto subMean =
            containerMat - Eigen::VectorXd::Ones(container.size()) * mean;
        double Var = subMean.transpose() * subMean;
        double dotProduct = containerMat.transpose() * containerMat; //
        tmpSum.emplace_back(dotProduct);
        tmpVar.emplace_back(Var / container.size()); // 方差
      }
      Eigen::Map<Eigen::VectorXd> tmpMat(tmpSum.data(), tmpSum.size());
      Eigen::MatrixXd::Index p, q;
      auto minDiff = tmpMat.minCoeff(&p, &q);
      auto minVar = tmpVar[p];
      // if (minVar > 0.002) {
      //   continue;
      // }

      std::pair<int, int> umatch;
      umatch.first = match.first;
      umatch.second = match.second[p];

      cam2mapMatch.emplace_back(umatch);
      continue;
    }
    cam2mapMatch.emplace_back(
      std::pair<int, int>(match.first, match.second[0]));
  }

  std::vector<std::pair<int, int>> cam2mapMatchPro;
  std::vector<double> backThress, backThressPro;

  for (auto &match : cam2mapMatch) {
    double backThres = 40;
    auto &camLine = camLaneRaw.lane_markers[match.first];
    auto &mapLine = mapLanes[match.second].second;

    auto &&end = camLine.end < 25 ? camLine.end : 25;
    backThres = backThres < camLine.end ? backThres : camLine.end;
    auto lower = std::lower_bound(mapLine.begin(), mapLine.end(), end,
                                  comp<Eigen::Vector3d>);
    std::cout << "end distance: " << end << std::endl;
    if (lower == mapLine.end() || lower == mapLine.begin() || lower->x() < 0) {
      std::cout << "找不到满足条件的点" << std::endl;
      cam2mapMatchPro.emplace_back(match);
      backThress.emplace_back(backThres);
      continue;
    }
    Eigen::Vector3d backPt = *(lower - 1);
    std::cout << "backPt: " << backPt.transpose() << std::endl;
    double x = backPt.x();
    double y = camLine.c0 + camLine.c1 * x + camLine.c2 * x * x +
               camLine.c3 * x * x * x;
    double backDiff = fabs(y - backPt.y());
    backThres -= 15 * backDiff;

    if (backDiff / (end - x) > 0.7 ||
        backDiff > 1) {
      std::cout << "backDiff: " << backDiff
                << ", back c1 approximate: " << (backDiff / x)
                << ", 此次删除的匹配是: " << match.first << " : "
                << match.second << std::endl;
      continue;
    }
    backThress.emplace_back(backThres);
    cam2mapMatchPro.emplace_back(match);
  }
  cam2mapMatch = cam2mapMatchPro;
  cam2mapMatchPro.clear();
  std::cout << "back点校验后剩余的匹配数: " << cam2mapMatch.size() << std::endl;


  double refDiff = 1e10;
  for (auto &onePair : cam2mapMatch) {
    double diff = fabs(mapc0s[onePair.second] - camc0s[onePair.first]);
    refDiff = refDiff < diff ? refDiff : diff;
  }
  for (int i = 0; i < cam2mapMatch.size(); i++) {
    auto &match = cam2mapMatch[i];
    double c0diff = fabs(mapc0s[match.second] - camc0s[match.first]);
    std::cout << c0diff << std::endl;
    if (c0diff <= refDiff + 0.6) {
      cam2mapMatchPro.emplace_back(match);
      backThressPro.emplace_back(backThress[i]);
    }
  }

  if (!cam2mapMatchPro.empty()) {
    cam2mapMatch = cam2mapMatchPro;
    backThress = backThressPro;
  }
  cam2mapMatchPro.clear();
  backThressPro.clear();


  for (auto &match : cam2mapMatch) {
    std::cout << "c0 diff: " << mapc0s[match.second] << " : "
              << camc0s[match.first] << std::endl;
  }
  std::unordered_map<int, int> c2mmatch;
  for (auto &match : cam2mapMatch) {
    c2mmatch[match.first] = match.second;
  }

  std::vector<Eigen::Vector3d> camvaildc, mapvaildc;

  for (int j = 0; j < cam2mapMatch.size(); j++) {
    auto &oneLane = camLaneProed.lane_markers[cam2mapMatch[j].first];
    auto &mapLane = mapLanes[cam2mapMatch[j].second];
    if (oneLane.end == 0) {
      continue;
    }
    auto lower = std::lower_bound(mapLane.second.begin(), mapLane.second.end(),
                                  0, comp<Eigen::Vector3d>);
    Eigen::Vector2d xvals, yvals;
    xvals << lower->x(), (lower - 1)->x();
    yvals << lower->y(), (lower - 1)->y();

    Eigen::Vector2d poly = PolyFit(xvals, yvals);

    double fy = poly.x();
    Eigen::Vector3d camOrigin(0, oneLane.c0, 0), mapOrigin(0, fy, 0);

    mapvaildc.emplace_back(mapOrigin);
    camvaildc.emplace_back(camOrigin);
    mapPtsVehMatchc.emplace_back(mapOrigin);
    mapPtsUTMMatchc.emplace_back(basePose.quat_ * mapOrigin +
                                 basePose.utmPosition_);
    j2PtsVehMatchc.emplace_back(camOrigin);
    j2PtsUTMMatchc.emplace_back(basePose.quat_ * camOrigin +
                                basePose.utmPosition_);
    // std::cout << "have lane view range" << std::endl;
    Eigen::Vector3d minPt = Eigen::Vector3d::Zero();
    double start = oneLane.start;
    double end = oneLane.end;

    for (int i = 0; i < mapLane.second.size(); i++) {
      auto &mapPtVeh = mapLane.second[i];
      double x = mapPtVeh.x();
      double y = oneLane.c0 + oneLane.c1 * x + oneLane.c2 * x * x +
                 oneLane.c3 * x * x * x;

      minPt.x() = x;
      minPt.y() = y;

      double Thres = backThress[j];
      if (minPt.x() <= end && minPt.x() >= start && minPt.x() <= Thres &&
          fabs(minPt.y()) <= 10. && fabs(mapPtVeh.y()) <= 10.) {
        mapPtsVehMatchc.emplace_back(mapPtVeh);
        mapPtsUTMMatchc.emplace_back(basePose.quat_ * mapPtVeh +
                                     basePose.utmPosition_);
        j2PtsVehMatchc.emplace_back(minPt);
        j2PtsUTMMatchc.emplace_back(basePose.quat_ * minPt +
                                    basePose.utmPosition_);
        camvaildc.emplace_back(minPt);
        mapvaildc.emplace_back(mapPtVeh);
      }
    }

    if (camvaildc.size() > 0) {
      camvaild.emplace_back(std::pair<int, std::vector<Eigen::Vector3d>>(
        cam2mapMatch[j].first, camvaildc));
      mapvaild.emplace_back(std::pair<int, std::vector<Eigen::Vector3d>>(
        mapLane.first, mapvaildc));
      matches.emplace_back(
        std::pair<int, int>(cam2mapMatch[j].first, mapLane.first));
      mapPtsUTMMatch.insert(mapPtsUTMMatch.end(), mapPtsUTMMatchc.begin(),
                            mapPtsUTMMatchc.end());
      mapPtsVehMatch.insert(mapPtsVehMatch.end(), mapPtsVehMatchc.begin(),
                            mapPtsVehMatchc.end());
      j2PtsUTMMatch.insert(j2PtsUTMMatch.end(), j2PtsUTMMatchc.begin(),
                           j2PtsUTMMatchc.end());
      j2PtsVehMatch.insert(j2PtsVehMatch.end(), j2PtsVehMatchc.begin(),
                           j2PtsVehMatchc.end());
    }
    camvaildc.clear();
    mapvaildc.clear();
    mapPtsUTMMatchc.clear();
    mapPtsVehMatchc.clear();
    j2PtsUTMMatchc.clear();
    j2PtsVehMatchc.clear();
  }

  j2PtsVehDraw.clear();
  j2PtsUTMDraw.clear();

  for (auto &camLine : camLaneRaw.lane_markers) {
    for (int i = camLine.start; i < camLine.end; i += 5) {
      Eigen::Vector3d minPt = Eigen::Vector3d::Zero();
      double x = i;
      double y = camLine.c0 + camLine.c1 * x + camLine.c2 * x * x +
                 camLine.c3 * x * x * x;
      minPt.x() = x;
      minPt.y() = y;
      j2PtsVehDraw.emplace_back(minPt);
      j2PtsUTMDraw.emplace_back(basePose.quat_ * minPt + basePose.utmPosition_);
    }
  }
}



bool SyncData::GetFrame(MatchFrame &MixData) {
  if (!(mapInited_ & j2Inited_ & egoInited_)) {
    std::cerr << "Map || Ego || J2 Not Inited, Skipped" << std::endl;
    return false;
  }

  J2Data j2Msg = j2Msg_;
  double j2Stamp = j2Msg.stamp_;
  Localization::PointLoc::InsInfo nearestEgo;
  GetNearestEgo(j2Stamp, nearestEgo);

  MapSwitched nearestMap;
  GetNearestMap(j2Stamp, nearestMap);

  std::vector<Eigen::Vector3d> mapPtsUTM, mapPtsVeh;
  std::vector<std::pair<int, std::vector<Eigen::Vector3d>>> mapLanes;
  std::vector<double> mapc1s, mapc0s;

  SwitchMapLaneInfo(nearestMap, nearestEgo, mapPtsUTM, mapPtsVeh, mapLanes,
                    mapc1s, mapc0s);

  if (mapLanes.size() == 0 || mapLanes.size() != mapc1s.size()) {
    std::cerr << "mapLanes.size() == 0 or mapc1s.size() != mapLanes.size()"
              << std::endl;
    return false;
  }

  std::vector<Eigen::Vector3d> mapPtsSelectedVeh, mapPtsSelectedUTM,
    j2PtsSelectedVeh, j2PtsSelectedUTM, j2PtsVehDraw, j2PtsUTMDraw;
  std::vector<std::pair<int, std::vector<Eigen::Vector3d>>> camvaild, mapvaild;

  std::vector<std::pair<int, int>> matches;

  SwitchCamLaneInfoByMap(j2Msg, mapLanes, mapc1s, mapc0s, mapPtsVeh, nearestEgo,
                         j2PtsVehDraw, j2PtsUTMDraw, mapPtsSelectedVeh,
                         mapPtsSelectedUTM, j2PtsSelectedVeh, j2PtsSelectedUTM,
                         camvaild, mapvaild, matches);

  if (camvaild.size() == 0 || camvaild.size() != mapvaild.size()) {
    std::cerr << "SwitchCamLaneInfoByMap() 出错. J2处理后数据为0或大小不匹配。"
              << std::endl;
    return false;
  }

  LaneStatus status = LaneStatus::GOOD;
  if (camvaild.size() == 1) {
    status = LaneStatus::SINGLE;
  } else {
    std::vector<double> remotediff;
    for (int i = 0; i < camvaild.size(); i++) {
      remotediff.emplace_back(
        fabs(camvaild[i].second.back().y() - mapvaild[i].second.back().y()));
    }
    sort(remotediff.begin(), remotediff.end());
    if (remotediff.back() - remotediff.front() > 0.8) {
      status = LaneStatus::BAD;
    }
  }

  MixData.stamp = j2Stamp;
  MixData.Egmo = nearestEgo;
  MixData.Relative = nearestEgo;

  MixData.VcsLines.lanestatus = status;
  MixData.VcsLines.matchedIds = matches;
  MixData.VcsLines.mapLines = mapvaild;
  MixData.VcsLines.camLines = camvaild;

  MixData.GcsLines.lanestatus = status;
  MixData.GcsLines.matchedIds = matches;
  std::vector<std::pair<int, std::vector<Eigen::Vector3d>>> mapvalidUTM,
    camvalidUTM;

  for (int i = 0; i < matches.size(); i++) {
    auto camLane = camvaild[i];
    auto mapLane = mapvaild[i];
    for (int j = 0; j < camLane.second.size(); j++) {
      camLane.second[j] =
          nearestEgo.quat_ * camLane.second[j] + nearestEgo.utmPosition_;
      mapLane.second[j] =
          nearestEgo.quat_ * mapLane.second[j] + nearestEgo.utmPosition_;
    }
    mapvalidUTM.emplace_back(mapLane);
    camvalidUTM.emplace_back(camLane);
  }
  MixData.GcsLines.mapLines = mapvalidUTM;
  MixData.GcsLines.camLines = camvalidUTM;

  MixData.VcsPoints.sensorMatchedPts = j2PtsSelectedVeh;
  MixData.VcsPoints.mapMatchedPts = mapPtsSelectedVeh;
  MixData.VcsPoints.ptNums = mapPtsSelectedVeh.size();

  MixData.GcsPoints.sensorMatchedPts = j2PtsSelectedUTM;
  MixData.GcsPoints.mapMatchedPts = mapPtsSelectedUTM;
  MixData.GcsPoints.ptNums = mapPtsSelectedUTM.size();

  // to draw
  MixData.j2UTMDraw = j2PtsUTMDraw;
  MixData.j2VehDraw = j2PtsVehDraw;
  MixData.mapUTMDraw = mapPtsUTM;
  MixData.mapVehDraw = mapPtsVeh;

  return true;
}

bool SyncData::GetOutput(MatchFrame &MixData) {
  sycmtx_.lock();
  if (synced_.size() == 0) {
    sycmtx_.unlock();
//      std::cout << "buffer_ 内不包含待处理数据, 请检查sync_ init情况"<< std::endl;
    return false;
  }
  MixData = synced_.back();
  synced_.pop_front();
  sycmtx_.unlock();
  return true;
}

bool SyncData::GetNearestMap(double &stamp, MapSwitched &NearestMap) {
  mapmtx_.lock();
  std::deque<MapSwitched> mapMsgBuffer = mapMsgBuffer_;
  mapmtx_.unlock();
  auto tmp = std::lower_bound(mapMsgBuffer.begin(), mapMsgBuffer.end(), stamp,
                              comp<MapSwitched>);
  if (tmp != mapMsgBuffer.end()) {
    if (tmp - mapMsgBuffer.begin() == 0 ||
        fabs((*tmp).stamp_ - stamp) < fabs((*(tmp - 1)).stamp_ - stamp)) {
      NearestMap = *tmp;
    } else {
      NearestMap = *(tmp - 1);
    }
    return 1;
  } else {
    std::cout << "use buffer back as nearest map" << std::endl;
    NearestMap = mapMsgBuffer.back();
    return false;
  }
}

bool SyncData::GetNearestEgo(double &stamp,Localization::PointLoc::InsInfo &NearestEgo) {
  egomtx_.lock();
  std::deque<Localization::PointLoc::InsInfo> egoMsgBuffer = egoMsgBuffer_;
  egomtx_.unlock();
  auto tmp = std::lower_bound(egoMsgBuffer.begin(), egoMsgBuffer.end(), stamp,
                              comp<Localization::PointLoc::InsInfo>);
  if (tmp != egoMsgBuffer.end()) {
    if (tmp - egoMsgBuffer.begin() == 0 ||
        fabs((*tmp).stamp_ - stamp) < fabs((*(tmp - 1)).stamp_ - stamp)) {
      NearestEgo = *tmp;
    } else {
      NearestEgo = *(tmp - 1);
    }
    return 1;
  } else {
    std::cout << "use buffer back as nearest ego" << std::endl;
    NearestEgo = egoMsgBuffer.back();
    return false;
  }
}

void SyncData::Interpolate3DBetweenTwoPointsByMeter(
    const Eigen::Vector3d &a, const Eigen::Vector3d &b, const double &step,
    std::vector<Eigen::Vector3d> &res) {
  res.clear();
  const Eigen::Vector3d ab = b - a;
  if (ab.norm() < step || ab.norm() > 150) {
    return;
  }
  res.emplace_back(a);

  const Eigen::Vector3d abNorm = ab.normalized();
  const double lengthnorm = ab.norm();

  const double lend = lengthnorm / step;
  const int len = static_cast<int>(lend);
  for (int i = 1; i < len; i++) {
    res.emplace_back(a + static_cast<double>(i) * step * abNorm);
  }
}
} // namespace PointLoc
} // namespace Localization
